import os
import subprocess
import threading


from .printer import console

class ZigToolchain:
    def __init__(self, language: str, cache_dir: str):
        self.compiler = self.__get_compiler(language)
        self.cache_dir = cache_dir

        os.environ['ZIG_LOCAL_CACHE_DIR'] = self.cache_dir

    def compile(self,
            output: str,
            input_file: str,
            c_flags: list[str] = [],
        ) -> threading.Thread:
        if not input_file:
            print("Fatal Error: No input file")
            raise SystemExit
        compile_cmd = (['zig', self.__get_compiler(os.path.splitext(input_file)[1])] + c_flags + ['-o', output, '-c'] + [input_file])
        console.print('[ Compiling ]', os.path.basename(input_file))
        # console.print('[ command ]', ' '.join(compile_cmd))
        return threading.Thread(target=self.__rc, args=(compile_cmd,))

    def link(self,
            output: str,
            input_files: list[str],
            l_flags: list[str] = []
        ) -> None:
        if not input_files:
            print("Fatal Error: No input files")
            raise SystemExit
        link_cmd = (['zig', self.compiler] + input_files + l_flags + ['-o', output])
        console.print('[ Linking ]', len(input_files), 'files')
        # console.print('[ command ]', ' '.join(link_cmd))
        self.__rc(link_cmd)

    def archive(self,
            output: str,
            input_files: list[str],
            options: list[str]
        ) -> None:
        if not input_files:
            print("Fatal Error: No input files")
            raise SystemExit
        archive_cmd = (['zig', 'build-lib', '-lc', '-lstdc++'] + input_files + options + ['-femit-bin=' + output])
        console.print('[ Archiving ]', os.path.basename(output))
        # console.print('[ command ]', ' '.join(archive_cmd))
        self.__rc(archive_cmd)

    def __rc(self, cmd) -> None:
        try:
            subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        except subprocess.CalledProcessError as e:
            console.print("Error:", e.stderr.decode())

    def __get_compiler(self, determiner: str) -> str:
        return 'cc' if determiner == '.c' else 'c++'


import os
import io
import re
import sys
import shutil
import json
import pkgconfig
import subprocess


from . import C, PROJECT, LOCAL_INSTALL_DIR, PROJECT_ROOT, NATIVE
from .printer import Print_Error


def Run_Command(*cmd, dir: str = PROJECT_ROOT, shell: bool = False, capture_output: bool = False) -> subprocess.CompletedProcess:
    try:
        result = subprocess.run(
            list(cmd),
            cwd=dir,
            shell=shell,
            capture_output=capture_output,
            text=True,
            check=True
        )
        return result
    except subprocess.CalledProcessError as e:
        Print_Error(f"Command failed with return code {e.returncode}: {e.cmd}")
        raise SystemError


# Finds the pkgconfig file of the library and returns its values.
def Grab_Dependency(
        library_name: str,
        debug: bool = False,
        triplet: str = NATIVE,
        dependency_dir: str = None
    ) -> (dict | None):
    dependency_dir = os.path.join(LOCAL_INSTALL_DIR, triplet) if dependency_dir is None else dependency_dir
    os.environ['PKG_CONFIG_PATH'] = f'{dependency_dir}{'/debug' if debug else ''}/lib/pkgconfig'
    return {
        'name': library_name,
        'debug': debug,
        'cflags': pkgconfig.cflags(library_name),
        'libs': pkgconfig.libs(library_name)
    } if pkgconfig.exists(library_name) else None

# Returns a library name based on the type of library and operating system.
def Generate_Library_Name(library_name: str, library_type: str, os_triplet: str) -> str:
    os_triplet = os_triplet.lower()
    if "darwin" in os_triplet or "macos" in os_triplet:
        os_prefix = 'lib'
        if library_type == 'static':
            os_suffix = '.a'
        elif library_type == 'dynamic':
            os_suffix = '.dylib'
        else:
            raise ValueError('Unsupported library type')
    elif "windows" in os_triplet:
        os_prefix = ''
        if library_type == 'static':
            os_suffix = '.lib'
        elif library_type == 'dynamic':
            os_suffix = '.dll'
        else:
            raise ValueError('Unsupported library type')
    else:
        os_prefix = 'lib'
        if library_type == 'static':
            os_suffix = '.a'
        elif library_type == 'dynamic':
            os_suffix = '.so'
        else:
            raise ValueError('Unsupported library type')

    return f'{os_prefix}{library_name}{os_suffix}'

# Checks if package is installed
def Check_Package_Installed(package_name: str) -> bool:
    vcpkg_info_dir = './.candie/packages/vcpkg/info'
    if  os.path.isdir(vcpkg_info_dir):
        if package_name in (p.split('_')[0] for p in os.listdir(vcpkg_info_dir)):
            return True
    return False

# Grabs the files files with the provided extension from a directory
def Grab_Files(directory_path: str, file_ext: str) -> tuple[str, ...]:  
    source_directory_path = os.path.abspath(directory_path)
    return tuple(
        str(file.path) for file in os.scandir(source_directory_path) 
        if file.is_file() and file.name.endswith(file_ext)
    )

# Grabs the source (.c , .cpp) files from the specified directory
def Grab_Sources(source_directory_path: str) -> tuple[str, ...]:
    if PROJECT['language'] == C:
        return Grab_Files(source_directory_path, '.c')
    else:
        return Grab_Files(source_directory_path, '.c') + Grab_Files(source_directory_path, '.cpp')

# Grabs the header (.h, .hpp) files from the specified directory
def Grab_Headers(source_directory_path: str) -> tuple[str, ...]:
    if PROJECT['language'] == C:
        return Grab_Files(source_directory_path, '.h')
    else:
        return Grab_Files(source_directory_path, '.h') + Grab_Files(source_directory_path, '.hpp')

# Copies directory files from source to destination direcory
def Copy_Directory(src_dir, dest_dir) -> list[str]:
    copied: list[str] = []
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    for root, _, files in os.walk(src_dir):
        rel_path = os.path.relpath(root, src_dir)
        dest_subdir = os.path.join(dest_dir, rel_path)
        if not os.path.exists(dest_subdir):
            os.makedirs(dest_subdir)
            copied.append(dest_subdir)
        for file in files:
            src_file = os.path.join(root, file)
            dest_file = os.path.join(dest_subdir, file)
            shutil.copy2(src_file, dest_file)
            copied.append(dest_file)
    return copied

# Copy pastes files.
def Copy_Files(file_paths, destination_dir: str):
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)
    for file_path in file_paths:
        file_name = os.path.basename(file_path)
        destination_file_path = os.path.join(destination_dir, file_name)
        try:
            shutil.copy(file_path, destination_file_path)
            print(f"Copied {file_path} to {destination_file_path}")
        except Exception as e:
            print(f"Error copying {file_path}: {str(e)}")

def Generate_PkgConfig_File(
        name: str,
        description: str,
        version: str,
        url: str,
        requires: list[str],
        libs: list[str],
        cflags: list[str],
        debug: bool = False
    ):
    return f'''
prefix=${{pcfiledir}}/../..{'/..' if debug else ''}
exec_prefix=${{prefix}}
includedir=${{prefix}}/include
libdir=${{prefix}}/{'debug/lib' if debug else 'lib'}

Name: {name}
Description: {description}
Version: {version}
URL: {url}

Libs: "-L${{libdir}}" {' '.join(libs)}
Requires: {' '.join(requires)}
Cflags: "-I${{includedir}}" {' '.join(cflags)}
'''


import re
import os
import io
import platform

_GLOBAL_NAMESPACE_ = globals()

PROJECT_ROOT = os.getcwd()
ARCHITECTURE = platform.architecture()
SYSTEM = platform.system()

# Files
CANDIEFILE = f"{PROJECT_ROOT}/Candie.build"

# Directories
CANDIE_DIR = f"{PROJECT_ROOT}/.candie"
LOCAL_INSTALL_DIR = f"{CANDIE_DIR}/installed"
CACHE_DIR = f"{CANDIE_DIR}/cache"

# Project object
PROJECT: dict = {}

# Languages
C:      str = '.c'
CPP:    str = '.cpp'

# Library types
STATIC = 'static'
SHARED = 'dynamic'

# Triplets
# 32 bit ARM is not working for some reason :(
TRIPLETS = [
    (NATIVE := 'native'),

    (X86_WINDOWS    := 'x86-windows'),
    (X64_WINDOWS    := 'x86_64-windows'),
    # (ARM_WINDOWS  := 'arm-windows'),
    (ARM64_WINDOWS  := 'aarch64-windows'),

    (X86_LINUX      := 'x86-linux'),
    (X64_LINUX      := 'x86_64-linux'),
    # (ARM_LINUX    := 'arm-linux'),
    (ARM64_LINUX    := 'aarch64-linux'),

    # (ARM_MACOS    := 'arm-macos'),
    (ARM64_MACOS    := 'aarch64-macos'),
    (X86_MACOS      := 'x86-macos'),
    (X64_MACOS      := 'x86_64-macos')
]

# Compiler Args
C_ARGS = [

    # Optimization
    (O_0 := '-O0'),
    (O_1 := '-O1'),
    (O_2 := '-O2'),
    (O_3 := '-O3'),
    (O_S := '-Os'),
    (O_Z := '-Oz'),

    # Warning and error
    (W_ALL      := '-Wall'),
    (W_EXTRA    := '-Wextra'),
    (W_ERROR    := '-Werror'),
    (W_PEDANTIC := '-Wpedantic'),

    # Standard version c
    (STD_C89     := '-std=c89'),
    (STD_C90     := '-std=c90'),
    (STD_GNU89   := '-std=gnu89'),
    (STD_C99     := '-std=c99'),
    (STD_GNU99   := '-std=gnu99'),
    (STD_C11     := '-std=c11'),
    (STD_GNU11   := '-std=gnu11'),
    (STD_C17     := '-std=c17'),
    (STD_GNU17   := '-std=gnu17'),
    (STD_C23     := '-std=c23'),
    (STD_GNU23   := '-std=gnu23'),

    # Standard version c++
    (STD_CPP98   := '-std=c++98'),
    (STD_GNUPP98 := '-std=gnu++98'),
    (STD_CPP03   := '-std=c++03'),
    (STD_GNUPP03 := '-std=gnu++03'),
    (STD_CPP11   := '-std=c++11'),
    (STD_GNUPP11 := '-std=gnu++11'),
    (STD_CPP014  := '-std=c++14'),
    (STD_GNUPP14 := '-std=gnu++14'),
    (STD_CPP17   := '-std=c++17'),
    (STD_GNUPP17 := '-std=gnu++17'),
    (STD_CPP20   := '-std=c++20'),
    (STD_GNUPP20 := '-std=gnu++20'),
    (STD_CPP23   := '-std=c++23'),
    (STD_GNUPP23 := '-std=gnu++23'),
]

from .utils import (
    Grab_Files,
    Grab_Sources,
    Grab_Headers,
    Grab_Dependency,
)

from .genrators import Executable, Package, Library
from .compiler import ZigToolchain as Compiler
from .printer import Print_Error

grab_files       = Grab_Files
grab_sources     = Grab_Sources
grab_headers     = Grab_Headers
grab_dependency  = Grab_Dependency

compiler         = Compiler
executable       = Executable
library          = Library
package          = Package

def project(name: str, language: str, version: str = '1.0.0') -> None:
    PROJECT['name'] = name
    PROJECT['language'] = language
    PROJECT['version'] = version
    PROJECT['compiler'] = Compiler(PROJECT['language'], CACHE_DIR)

def parse(file: io.TextIOWrapper) -> dict:
    data = file.read()
    sections = re.findall(r'^\[(.*?)\](.*?)(?=^\[|\Z)', data, re.DOTALL | re.MULTILINE)
    parsed_data = {section.strip(): content.strip() for section, content in sections}
    return parsed_data

def setup() -> dict:
    if not os.path.isfile(CANDIEFILE):
        Print_Error("Build instruction file not found.")
        raise SystemExit
    
    for dir in [
        CANDIE_DIR,
        CACHE_DIR
    ]:
        if not os.path.isdir(dir):
            os.makedirs(dir)
    
    with open(CANDIEFILE, 'r') as f:
        sections = parse(f)
    
    exec(sections["SETUP"], _GLOBAL_NAMESPACE_)
    del sections["SETUP"]
    
    # console.print("Sections:", sections)
    # console.print("Project:", PROJECT)
    
    return sections

def execute(section: str) -> None:
    sections = setup()
    exec(sections[section], _GLOBAL_NAMESPACE_)


import os

from pathlib import Path


from . import STATIC, SHARED, PROJECT, LOCAL_INSTALL_DIR, TRIPLETS, NATIVE
from .utils import Generate_PkgConfig_File, Copy_Files, Grab_Dependency, Generate_Library_Name
from .printer import Print_Error


class Executable:
    def __init__(self,
            name: str, 
            *sources: str,
            dependencies: list[dict] = [],
            c_args: list[str] = [],
            link_args: list[str] = [],
            triplet = NATIVE,
            make_out_dir: str = '.',
        ):
        self.name = name
        self.sources = {str(Path(src).resolve()) for src in sources}

        # Dependencies 
        self.dependencies = dependencies

        # Compiler arguments
        self.cflags: list[str] = c_args
        self.lflags: list[str] = link_args

        # Triplet
        if triplet not in TRIPLETS:
            Print_Error(f"Invalid Triplet. ({triplet})")
            raise SystemExit
        self.triplet = triplet
        
        # Directories
        self.make_out_dir = make_out_dir
        if not os.path.isdir(make_out_dir):
            os.makedirs(make_out_dir)
    
    def link_against(self, dependency: dict):
        if dependency:
            self.cflags.append(dependency.get('cflags', ''))
            self.lflags.append(dependency.get('libs', ''))
        else:
            print("Error: invalid dependency")
        return self

    def create(self, debug: bool = False):
        # Setting triplet
        self.cflags.append('-target')
        self.cflags.append(self.triplet)

        # Add debugging symbols
        if debug: self.cflags.append('-g') 

        # Dependencies
        for dependency in self.dependencies:
            self.link_against(Grab_Dependency(dependency, debug=debug, triplet=self.triplet))

        # Build the Executable
        PROJECT['compiler'].link(
            output=f"{self.make_out_dir}/{self.name}",
            input_files=list(self.sources),
            l_flags=self.cflags + [lflag for lflags in self.lflags for lflag in lflags.split()]
        )


class Library:
    def __init__(self, 
            name: str, 
            lib_type: str, 
            *sources: str,
            dependencies: list[dict] = [],
            c_args: list[str] = [],
            triplet = NATIVE,
            make_out_dir: str = '.',
        ):
        self.name = name
        self.sources = {str(Path(src).resolve()) for src in sources}
        self.lib_type = STATIC if lib_type not in [STATIC, SHARED] else lib_type

        # Triplet
        if triplet not in TRIPLETS:
            Print_Error(f"Invalid Triplet. ({triplet})")
            raise SystemExit
        self.triplet = triplet

        # Output directory
        self.make_out_dir = make_out_dir

        # Dependencies
        self.requirements: list[str] = []
        for dependency in dependencies:
            self.requirements.append(dependency.get('name', ''))
            self.link_against(dependency)

        # Compiler arguments
        self.cflags: list[str] = c_args

    def link_against(self, dependency: dict):
        if dependency:
            self.requirements.append(dependency.get('name', ''))
            self.cflags.append(dependency.get('cflags', ''))
            self.lflags.append(dependency.get('libs', ''))
        else:
            print("Error: invalid dependency")
        return self

    def create(self, debug: bool = False):
        libdir = os.path.join(self.make_out_dir, 'debug', 'lib') if debug else os.path.join(self.make_out_dir, 'lib')
        if not os.path.isdir(libdir):
            os.makedirs(libdir)
                
        # output path
        output_path = os.path.join(libdir, Generate_Library_Name(self.name, self.lib_type, self.triplet))

        # setup target
        self.cflags.append('-target')
        self.cflags.append(self.triplet)

        # Strip debugging symbols
        if not debug: self.cflags.append('-fstrip')

        # Create library file
        if self.lib_type == STATIC:
            PROJECT['compiler'].archive(
                output=output_path,
                input_files=list(self.sources),
                options=self.cflags
            )
        elif self.lib_type == SHARED:
            self.lflags.append('-shared')
            PROJECT['compiler'].link(
                output=output_path,
                input_files=list(self.sources),
                l_flags=self.cflags
            )
        return self


class Package:
    def __init__(self, 
            name: str,
            *libraries: Library,
            triplet: str = NATIVE,
            install_dir: str = LOCAL_INSTALL_DIR,
            cflags: list[str] = [],
            description: str = '',
            version: str = '0.0.0',
            url: str = ''
        ):
        self.name = name
        self.description = description
        self.version = version
        self.url = url
        self.cflags = cflags

        # Triplet
        if triplet not in TRIPLETS:
            Print_Error(f"Invalid Triplet. ({triplet})")
            raise SystemExit
        self.triplet = triplet

        # Directories
        self.install_dir = os.path.join(install_dir, self.triplet)
        if not os.path.isdir(self.install_dir):
            os.makedirs(self.install_dir)

        # Libraries
        self.libraries = libraries
    
    def install_headers(self, *headers: str):
        Copy_Files(headers, f"{self.install_dir}/include")
        return self

    def create(self, debug: bool = False):
        pkgconfig_dir = f"{self.install_dir}{'/debug/lib/pkgconfig' if debug else '/lib/pkgconfig'}"
        if not os.path.isdir(pkgconfig_dir):
            os.makedirs(pkgconfig_dir)

        # Make libraries
        for library in self.libraries:
            library.make_out_dir = self.install_dir
            library.triplet = self.triplet
            library.create(debug=debug)
        
        # Write pkgconfig file
        with open(os.path.join(pkgconfig_dir, self.name + '.pc'), 'w') as dotpcfile:
            dotpcfile.write(Generate_PkgConfig_File(
                name=self.name,
                description=self.description,
                version=self.version,
                url=self.url,
                requires=[req for lib in self.libraries for req in lib.requirements],
                cflags=self.cflags,
                libs=[('-l' + lib.name) for lib in self.libraries],
                debug=debug
            ))



from rich.console import Console


console = Console()


def Print_Error(err: str):
    console.print('[ [bright_red]Error[/bright_red] ] ', err)


def Print_Warning(warn: str):
    console.print('[ [deep_pink2]Warning[/deep_pink2] ] ', warn)


def Log_Msg(msg: str):
    console.log('[ [chartreuse1]Msg[/chartreuse1] ] ', msg)